#include <windows.h>
#include "runtime.h"
#include "bit.h"
#include "smaf.h"

BOOL findchunk4(BYTE* data, DWORD dsize, char* cname, CHUNK* ch)
{
	DWORD cur;

	for(cur = 0; cur < dsize; )
	{
		if(!memcomp(data + cur, cname, 4))
		{
			ch->data = data + cur + 8;
			ch->size = SwapDword(data + cur + 4);
			return TRUE;
		}
		cur += 8 + SwapDword(data + cur + 4);
	}

	return FALSE;
}

BOOL findchunk3(BYTE* data, DWORD dsize, char* cname, CHUNK* ch)
{
	DWORD cur;

	for(cur = 0; cur < dsize; )
	{
		if(!memcomp(data + cur, cname, 3))
		{
			ch->data = data + cur + 8;
			ch->size = SwapDword(data + cur + 4);
			return TRUE;
		}
		cur += 8 + SwapDword(data + cur + 4);
	}

	return FALSE;
}

DWORD GetVValAndForward(BYTE** pdata)
{
	DWORD val = 0;

	do
	{
		val <<= 7;
		val |= **pdata & 0x7F;
	}while(*(*pdata)++ & 0x80);

	return val;
}

DWORD GetVValSize(BYTE* pdata)
{
	BYTE* p = pdata;

	do
	{
		;
	}while(*p++ & 0x80);

	return p - pdata;
}

DWORD GetVValAndForward4HPS(BYTE** pdata)
{
	DWORD val = 0;

	if(**pdata & 0x80)
		val = (((*(*pdata)++) & 0x7F) + 1) << 7;

	val |= *(*pdata)++;

	return val;
}

DWORD GetVVal(BYTE* pdata)
{
	DWORD val = 0;

	do
	{
		val <<= 7;
		val |= *pdata & 0x7F;
	}while(*pdata++ & 0x80);

	return val;
}

void ClearEvents(SMAF* smaf)
{
	EVENT* e = smaf->events;

	while(e)
	{
		EVENT* next = e->next;
		hfree(e->data);
		hfree(e);
		e = next;
	}

	smaf->events = NULL;

	return;
}

void InsertEvent(SMAF* smaf, EVENT* e)
{
	EVENT* ev = smaf->events;
	EVENT* enx;

	if(ev)
	{
		while(ev->next)
		{
			if(e->time < ev->next->time)
				break;
			ev = ev->next;
		}
	}

	if(!ev)
	{
		enx = NULL;
		smaf->events = e;
	}else
	{
		enx = ev->next;
		ev->next = e;
	}

	e->next = enx;

	return;
}

// for HPS: convert to MSF
void ConvertEvents2(SMAF* smaf, BYTE* data, DWORD size, BOOL fsetup, DWORD start, DWORD stop, UINT chshift)
{
	BYTE* p;
	DWORD tdur = 0;
	signed int octshift[4] = {0};

	BOOL started = FALSE;
	BOOL stoped = FALSE;

	for(p = data; p < data + size; )
	{
		EVENT* e;
		BYTE esize = 1;	// command size

		if(!fsetup)
		{
			if(!smaf->start && (DWORD)(p - data) == start)
				started = TRUE;
			if(!smaf->stop && (DWORD)(p - data) == stop)
				stoped = TRUE;
		}

		e = halloc(sizeof(EVENT));
		if(fsetup)
			e->duration = 0;
		else
			e->duration = GetVValAndForward4HPS(&p) * smaf->dbase;
		tdur += e->duration;
		e->time = tdur;
		e->status = 0;
		e->size = 0;
		e->data = NULL;

		// Parse status!!
		if(p[0] == 0xFF && p[1] == 0xF0)	// Exclusive
		{
			esize = 2 + 1 + p[2];
			// need exclusive convert
#if 1
			e->size = 1 + p[2];
			e->status = 0xF0;
			e->data = halloc(e->size);
			memcopy(e->data, p + 2, e->size);
#else
			hfree(e);
			p += esize;
			continue;
#endif
		}else if(p[0] == 0xFF && p[1] == 0x00)	// NOP
		{
			esize = 2;
			e->size = 1;
			e->data = halloc(e->size);
			e->status = 0xFF;
			e->data[0] = 0x00;
		}else
		{
			if(p[0] != 0x00)	// Note
			{
				BYTE ch;
				BYTE oct;
				BYTE note;
				BYTE absnote;

				ch = (p[0] >> 6 & 0x03);
				oct = p[0] >> 4 & 0x03;
				note = p[0] & 0x0F;
				// MIDI's 69 = 440Hz(A) = Oct2,9
				absnote = 36 + (oct + octshift[ch]) * 12 + note;

				e->status = 0x80 | (ch + chshift);
				{
					BYTE* q = p + 1;

					GetVValAndForward4HPS(&q);
					esize = q - p;
					{
						DWORD siz;
						siz = q - p;
						if(siz == 1 + 2 && p[1] == 0xFF)	// vval over 2byte.
							siz++;
						e->size = siz;
					}
					e->data = halloc(e->size);

					e->data[0] = absnote;
					if(p[1] & 0x80)
					{
						if(p[1] == 0xFF)
						{
							e->data[1] = 0x80;
							e->data[2] = p[1] + 1;
							e->data[3] = p[2];
						}else
						{
							e->data[1] = p[1] + 1;
							e->data[2] = p[2];
						}
					}else
						e->data[1] = p[1];
				}
			}else
			{
				BYTE ch;
				ch = (p[1] >> 6 & 0x03) + chshift;

				if(p[1] == 0x00 && p[2] == 0x00 && p[3] == 0x00)	// EOS?
					esize = 4;
				if((p[1] & 0x30) == 0x30)
				{
					if((p[1] & 0x0F) == 0x00)	// Program Change
					{
						esize = 3;
						e->size = 1;
						e->data = halloc(e->size);
						e->status = 0xC0 | ch;
						e->data[0] = p[2];
					}
					if((p[1] & 0x0F) == 0x01)	// Bank Select
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x00;	// banksel
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x02)	// Octave Shift
					{
						UINT ch = p[0] >> 6 & 0x03;

						esize = 3;
						if(0x01 <= p[2] && p[2] <= 0x04)
							octshift[ch] = p[2];
						if(0x81 <= p[2] && p[2] <= 0x84)
							octshift[ch] = -(p[2] - 0x80);

#if 0
						// NOP
						e->size = 1;
						e->data = halloc(e->size);
						e->status = 0xFF;
						e->data[0] = 0x00;
#endif
						hfree(e);
						p += esize;

						continue;
					}
					if((p[1] & 0x0F) == 0x03)	// Modulation
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x01;	// mod.
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x04)	// PitchBend
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xE0 | ch;
						e->data[0] = 0x00;
						e->data[1] = p[2];	// tenuki.
					}
					if((p[1] & 0x0F) == 0x07)	// Volume
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x07;	// vol
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x0A)	// Pan
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0A;	// pan
						e->data[1] = p[2];
					}
					if((p[1] & 0x0F) == 0x0B)	// Expression
					{
						esize = 3;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0B;	// exp.
						e->data[1] = p[2];
					}
				}else
				{
					if((p[1] & 0x30) == 0x00)	// Expression
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x0B;	// exp.
						e->data[1] = (p[1] & 0xF0) << 3;
					}
					if((p[1] & 0x30) == 0x10)	// PitchBend
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xE0 | ch;
						e->data[0] = 0x00;
						e->data[1] = (p[1] & 0xF0) << 3;	// chou tenuki
					}
					if((p[1] & 0x30) == 0x20)	// Modulation
					{
						esize = 2;
						e->size = 2;
						e->data = halloc(e->size);
						e->status = 0xB0 | ch;
						e->data[0] = 0x01;	// mod.
						e->data[1] = (p[1] & 0xF0) << 3;
					}
				}
			}
		}

		if(esize == 1)
		{
			TCHAR ts[256];

			wsprintf(ts, "What's this? %02X", p[0]);
			MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);
		}

		p += esize;

		InsertEvent(smaf, e);

		if(started)
		{
			smaf->start = e;
			started = FALSE;
		}
		if(stoped)
		{
			smaf->stop = e;
			stoped = FALSE;
		}
	}

	return;
}

// for MSF
void ConvertEvents(SMAF* smaf, BYTE* data, DWORD size, BOOL fsetup, DWORD start, DWORD stop)
{
	BYTE* p;
	DWORD tdur = 0;

	BOOL started = FALSE;
	BOOL stoped = FALSE;

	for(p = data; p < data + size; )
	{
		EVENT* e;

		if(!fsetup)
		{
			if((DWORD)(p - data) == start)
				started = TRUE;
			if((DWORD)(p - data) == stop)
				stoped = TRUE;
		}

		e = halloc(sizeof(EVENT));
		if(fsetup)
			e->duration = 0;
		else
			e->duration = GetVValAndForward(&p) * smaf->dbase;
		tdur += e->duration;
		e->time = tdur;
		e->status = *p++;
		e->size = 0;
		e->data = NULL;

		// Parse status!!
		if((e->status & 0xF0) == 0x80)	// Note w/o verocity
		{
			/*
			BYTE* q = p + 1;

			GetVValAndForward(&q);
			e->size = q - p;
			*/
			e->size = 1 + GetVValSize(p + 1);
		}
		if((e->status & 0xF0) == 0x90)	// Note w/ verocity
		{
			/*
			BYTE* q = p + 2;

			GetVValAndForward(&q);
			e->size = q - p;
			*/
			e->size = 2 + GetVValSize(p + 2);
		}
		if((e->status & 0xF0) == 0xA0)	// Resv 3byte
		{
			e->size = 3;
		}
		if((e->status & 0xF0) == 0xB0)	// Control Change
		{
			e->size = 2;
		}
		if((e->status & 0xF0) == 0xC0)	// Program Change
		{
			e->size = 1;
		}
		if((e->status & 0xF0) == 0xD0)	// Resv 2byte
		{
			e->size = 2;
		}
		if((e->status & 0xF0) == 0xE0)	// PitchBend
		{
			e->size = 2;
		}
		if(e->status == 0xF0)	// Exclusive
		{
			BYTE* q = p;

			e->size = GetVValAndForward(&q);
			e->size += (q - p);	// VVal size
		}
		if(e->status == 0xFF)	// NOP or EOS
		{
			if(*p == 0x00)
				e->size = 1;
			if(*p == 0x2F)
				e->size = 2;
		}

		// copy.
		if(e->size)
		{
			e->data = halloc(e->size);
			memcopy(e->data, p, e->size);
			p += e->size;
		}else
			MessageBox(NULL, "Bug or illegal", "What", MB_OK);

		InsertEvent(smaf, e);

		if(started)
		{
			smaf->start = e;
			started = FALSE;
		}
		if(stoped)
		{
			smaf->stop = e;
			stoped = FALSE;
		}
	}

	return;
}

SMAF* FreeSmaf(SMAF* smaf)
{
	if(!smaf)
	{
		MessageBox(NULL, "already smaf == NULL", "BUG?", MB_ICONERROR | MB_OK);
		return NULL;
	}

	if(smaf->data)
		hfree(smaf->data);

	ClearEvents(smaf);

	hfree(smaf);

	return NULL;
}

SMAF* LoadSmaf(LPCTSTR fname)
{
	SMAF* smaf = NULL;
	HANDLE hf = NULL;
	DWORD read;

	if(!(smaf = halloc(sizeof(SMAF))))
		goto Error;

	smaf->data = NULL;
	smaf->events = NULL;

	smaf->start = NULL;
	smaf->stop = NULL;

	if((hf = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ, NULL, OPEN_EXISTING, 0, NULL)) == INVALID_HANDLE_VALUE)
	{
		TCHAR ts[MAX_PATH + 80];

		wsprintf(ts, "Can't open %s", fname);
		MessageBox(NULL, ts, "Error", MB_ICONERROR | MB_OK);

		goto Error;
	}

	smaf->size = GetFileSize(hf, NULL);

	if(!(smaf->data = halloc(smaf->size)))
	{
		MessageBox(NULL, "Can't alloc memory", "Error", MB_ICONERROR | MB_OK);
		goto Error;
	}
	if(!ReadFile(hf, smaf->data, smaf->size, &read, NULL))
	{
		read = 0;
	}
	if(read < smaf->size)
	{
		MessageBox(NULL, "Couldn't read file.", "Error", MB_ICONERROR | MB_OK);
		goto Error;
	}

	CloseHandle(hf);

	{
		CHUNK chk;
		CHUNK mtr;
		DWORD chkplus;
		DWORD chshift;

		if(!findchunk4(smaf->data, smaf->size, "MMMD", &chk))
		{
			MessageBox(NULL, "This is not SMAF", "Error", MB_ICONERROR | MB_OK);
			goto Error;
		}

		chk.size -= 2;	// MMMD's CRC

		// MTRւ̑Ή
		for(chkplus = 0, chshift = 0; chkplus < chk.size; chshift += 4)
		{
			if(!findchunk3(chk.data + chkplus, chk.size, "MTR", &mtr))
				break;

			chkplus = (mtr.data - chk.data) + mtr.size;

			{
				CHUNK dummy;

				switch(mtr.data[2])
				{
				case 0x00: smaf->dbase = 1; break;
				case 0x01: smaf->dbase = 2; break;
				case 0x02: smaf->dbase = 4; break;
				case 0x03: smaf->dbase = 5; break;
				case 0x10: smaf->dbase = 10; break;
				case 0x11: smaf->dbase = 20; break;
				case 0x12: smaf->dbase = 40; break;
				case 0x13: smaf->dbase = 50; break;
				default: MessageBox(NULL, "Unknown duration timebase", "Error", MB_ICONERROR | MB_OK); goto Error;
				}

				switch(mtr.data[3])
				{
				case 0x00: smaf->gbase = 1; break;
				case 0x01: smaf->gbase = 2; break;
				case 0x02: smaf->gbase = 4; break;
				case 0x03: smaf->gbase = 5; break;
				case 0x10: smaf->gbase = 10; break;
				case 0x11: smaf->gbase = 20; break;
				case 0x12: smaf->gbase = 40; break;
				case 0x13: smaf->gbase = 50; break;
				default: MessageBox(NULL, "Unknown gatetime timebase", "Error", MB_ICONERROR | MB_OK); goto Error;
				}

				if(mtr.data[0] == 0x00)	// HPS
				{
					DWORD start, stop;

					mtr.data += 4 + 2;	// skip ChStatus

					// init start,stop.
					if(findchunk4(mtr.data, mtr.size, "Mtsq", &dummy))
					{
						start = 0;
						stop = mtr.size;
					}

					if(findchunk4(mtr.data, mtr.size, "Mtsu", &dummy))
						ConvertEvents2(smaf, dummy.data, dummy.size, TRUE, 0, 0, chshift);

					if(findchunk4(mtr.data, mtr.size, "MspI", &dummy))
					{
						BYTE* p = dummy.data;

						while(p < dummy.data + dummy.size)
						{
							if(*p == 's' && *(p + 1) == 't')
							{
								p += 3;
								start = *(DWORD*)p;
								p += 4;
							}
							if(*p == 's' && *(p + 1) == 'p')
							{
								p += 3;
								stop = *(DWORD*)p;
								p += 4;
							}
							if(*p != ',')
								break;
							p++;
						}
					}

					if(findchunk4(mtr.data, mtr.size, "Mtsq", &dummy))
						ConvertEvents2(smaf, dummy.data, dummy.size, FALSE, start, stop, chshift);
				}

				if(mtr.data[0] == 0x01 || mtr.data[0] == 0x02)	// MSF
				{
					DWORD start, stop;

					mtr.data += 4 + 16;

					// init start,stop.
					if(findchunk4(mtr.data, mtr.size, "Mtsq", &dummy))
					{
						start = 0;
						stop = mtr.size;
					}

					if(findchunk4(mtr.data, mtr.size, "Mtsu", &dummy))
						ConvertEvents(smaf, dummy.data, dummy.size, TRUE, 0, 0);

					if(findchunk4(mtr.data, mtr.size, "MspI", &dummy))
					{
						BYTE* p = dummy.data;

						while(p < dummy.data + dummy.size)
						{
							if(*p == 's' && *(p + 1) == 't')
							{
								p += 3;
								start = SwapDword(p);
								p += 4;
							}
							if(*p == 's' && *(p + 1) == 'p')
							{
								p += 3;
								stop = SwapDword(p);
								p += 4;
							}
							if(*p != ',')
								break;
							p++;
						}
					}

					if(findchunk4(mtr.data, mtr.size, "Mtsq", &dummy))
					{
						if(mtr.data[-20] == 0x01)
						{
							BYTE* data;

							data = halloc(SwapDword(dummy.data));
							huffmanDecode(data, dummy.data + 4, SwapDword(dummy.data), dummy.size);
							ConvertEvents(smaf, data, SwapDword(dummy.data), FALSE, start, stop);
							hfree(data);
						}
						if(mtr.data[-20] == 0x02)
							ConvertEvents(smaf, dummy.data, dummy.size, FALSE, start, stop);
					}
				}
			}
		}
	}

	return smaf;
Error:
	if(smaf)
	{
		FreeSmaf(smaf);
	}
	if(hf)
	{
		CloseHandle(hf);
	}

	return NULL;
}
